// Copyright 2021 Parallel Finance Developer.
// This file is part of Parallel Finance.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0

// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

use super::*;
use frame_support::{ensure, traits::Get};
use sp_core::H256;
use sp_io::hashing::keccak_256;
use sp_std::vec::Vec;

/// EIP2612 permit typehash.
pub const PERMIT_TYPEHASH: [u8; 32] = keccak256!(
    "Permit(address owner,address spender,uint256 value,uint256 nonce,uint256 deadline)"
);

/// EIP2612 permit domain used to compute an individualized domain separator.
const PERMIT_DOMAIN: [u8; 32] = keccak256!(
    "EIP712Domain(string name,string version,uint256 chainId,address verifyingContract)"
);

pub struct Eip2612<Runtime, Metadata, Instance = ()>(PhantomData<(Runtime, Metadata, Instance)>);

impl<Runtime, Metadata, Instance> Eip2612<Runtime, Metadata, Instance>
where
    Metadata: Erc20Metadata,
    Instance: InstanceToPrefix + 'static,
    Runtime: pallet_balances::Config<Instance> + pallet_evm::Config + pallet_timestamp::Config,
    Runtime::RuntimeCall: Dispatchable<PostInfo = PostDispatchInfo> + GetDispatchInfo,
    Runtime::RuntimeCall: From<pallet_balances::Call<Runtime, Instance>>,
    <Runtime::RuntimeCall as Dispatchable>::RuntimeOrigin: From<Option<Runtime::AccountId>>,
    BalanceOf<Runtime, Instance>: TryFrom<U256> + Into<U256>,
    <Runtime as pallet_timestamp::Config>::Moment: Into<U256>,
{
    fn compute_domain_separator(address: H160) -> [u8; 32] {
        let name: H256 = keccak_256(Metadata::name().as_bytes()).into();

        let version: H256 = keccak256!("1").into();

        let chain_id: U256 = Runtime::ChainId::get().into();

        let domain_separator_inner = EvmDataWriter::new()
            .write(H256::from(PERMIT_DOMAIN))
            .write(name)
            .write(version)
            .write(chain_id)
            .write(Address(address))
            .build();

        keccak_256(&domain_separator_inner)
    }

    pub fn generate_permit(
        address: H160,
        owner: H160,
        spender: H160,
        value: U256,
        nonce: U256,
        deadline: U256,
    ) -> [u8; 32] {
        let domain_separator = Self::compute_domain_separator(address);

        let permit_content = EvmDataWriter::new()
            .write(H256::from(PERMIT_TYPEHASH))
            .write(Address(owner))
            .write(Address(spender))
            .write(value)
            .write(nonce)
            .write(deadline)
            .build();
        let permit_content = keccak_256(&permit_content);

        let mut pre_digest = Vec::with_capacity(2 + 32 + 32);
        pre_digest.extend_from_slice(b"\x19\x01");
        pre_digest.extend_from_slice(&domain_separator);
        pre_digest.extend_from_slice(&permit_content);
        keccak_256(&pre_digest)
    }

    // Translated from
    // https://github.com/Uniswap/v2-core/blob/master/contracts/UniswapV2ERC20.sol#L81
    pub(crate) fn permit(handle: &mut impl PrecompileHandle) -> EvmResult<PrecompileOutput> {
        handle.record_cost(RuntimeHelper::<Runtime>::db_read_gas_cost())?;

        let mut input = EvmDataReader::new_skip_selector(handle.input())?;
        let owner: H160 = input.read::<Address>()?.into();
        let spender: H160 = input.read::<Address>()?.into();
        let value: U256 = input.read()?;
        let deadline: U256 = input.read()?;
        let v: u8 = input.read()?;
        let r: H256 = input.read()?;
        let s: H256 = input.read()?;

        // pallet_timestamp is in ms while Ethereum use second timestamps.
        let timestamp: U256 = (pallet_timestamp::Pallet::<Runtime>::get()).into() / 1000;

        ensure!(deadline >= timestamp, revert("permit expired"));

        let nonce = NoncesStorage::<Instance>::get(owner);

        let permit = Self::generate_permit(
            handle.context().address,
            owner,
            spender,
            value,
            nonce,
            deadline,
        );

        let mut sig = [0u8; 65];
        sig[0..32].copy_from_slice(r.as_bytes());
        sig[32..64].copy_from_slice(s.as_bytes());
        sig[64] = v;

        let signer = sp_io::crypto::secp256k1_ecdsa_recover(&sig, &permit)
            .map_err(|_| revert("invalid permit"))?;
        let signer = H160::from(H256::from_slice(keccak_256(&signer).as_slice()));

        ensure!(
            signer != H160::zero() && signer == owner,
            revert("invalid permit")
        );

        NoncesStorage::<Instance>::insert(owner, nonce + U256::one());

        {
            let amount =
                Erc20BalancesPrecompile::<Runtime, Metadata, Instance>::u256_to_amount(value)
                    .unwrap_or_else(|_| Bounded::max_value());

            let owner: Runtime::AccountId = Runtime::AddressMapping::into_account_id(owner);
            let spender: Runtime::AccountId = Runtime::AddressMapping::into_account_id(spender);
            ApprovesStorage::<Runtime, Instance>::insert(owner, spender, amount);
        }

        LogsBuilder::new(handle.context().address)
            .log3(
                SELECTOR_LOG_APPROVAL,
                owner,
                spender,
                EvmDataWriter::new().write(value).build(),
            )
            .record(handle)?;

        Ok(succeed([]))
    }

    pub(crate) fn nonces(handle: &mut impl PrecompileHandle) -> EvmResult<PrecompileOutput> {
        handle.record_cost(RuntimeHelper::<Runtime>::db_read_gas_cost())?;

        let mut input = EvmDataReader::new_skip_selector(handle.input())?;
        let owner: H160 = input.read::<Address>()?.into();

        let nonce = NoncesStorage::<Instance>::get(owner);

        Ok(succeed(EvmDataWriter::new().write(nonce).build()))
    }

    pub(crate) fn domain_separator(
        handle: &mut impl PrecompileHandle,
    ) -> EvmResult<PrecompileOutput> {
        handle.record_cost(RuntimeHelper::<Runtime>::db_read_gas_cost())?;

        let domain_separator: H256 =
            Self::compute_domain_separator(handle.context().address).into();

        Ok(succeed(
            EvmDataWriter::new().write(domain_separator).build(),
        ))
    }
}
